// Copyright (c) Microsoft Corporation. All rights reserved.
// Licensed under the MIT license. See LICENSE file in the project root for full license information.

using System;
using System.Collections.Generic;
using System.IO;

using Microsoft.VisualStudio.TestPlatform.MSTestAdapter.PlatformServices.Interface;
#if NETFRAMEWORK
using Microsoft.VisualStudio.TestPlatform.MSTestAdapter.PlatformServices.Utilities;
#endif
using Microsoft.VisualStudio.TestPlatform.ObjectModel;
using Microsoft.VisualStudio.TestPlatform.ObjectModel.Adapter;
using Microsoft.VisualStudio.TestPlatform.ObjectModel.Utilities;
using Microsoft.VisualStudio.TestPlatform.PlatformAbstractions.Interfaces;
using Microsoft.VisualStudio.TestTools.UnitTesting;

namespace Microsoft.VisualStudio.TestPlatform.MSTestAdapter.PlatformServices;

/// <summary>
/// A host that loads the test source. This can be in isolation for desktop using an AppDomain or just loading the source in the current context.
/// </summary>
public class TestSourceHost : ITestSourceHost
{
#if !WINDOWS_UWP
    private readonly string _sourceFileName;
    private string? _currentDirectory = null;
#endif

#if NETFRAMEWORK

    /// <summary>
    /// Determines whether child-appdomain needs to be created based on DisableAppDomain Flag set in runsettings.
    /// </summary>
#pragma warning disable SA1214 // Readonly fields should appear before non-readonly fields
    private readonly bool _isAppDomainCreationDisabled;
#pragma warning restore SA1214 // Readonly fields should appear before non-readonly fields

    private readonly IRunSettings? _runSettings;
    private readonly IFrameworkHandle? _frameworkHandle;
    private readonly IAppDomain _appDomain;

    /// <summary>
    /// Assembly resolver used in the current app-domain.
    /// </summary>
    private AssemblyResolver? _parentDomainAssemblyResolver;

    /// <summary>
    /// Assembly resolver used in the new child app-domain created for discovery/execution.
    /// </summary>
    private AssemblyResolver? _childDomainAssemblyResolver;

    private string? _targetFrameworkVersion;
#endif

    /// <summary>
    /// Initializes a new instance of the <see cref="TestSourceHost"/> class.
    /// </summary>
    /// <param name="sourceFileName"> The source file name. </param>
    /// <param name="runSettings"> The run-settings provided for this session. </param>
    /// <param name="frameworkHandle"> The handle to the test platform. </param>
    public TestSourceHost(string sourceFileName, IRunSettings? runSettings, IFrameworkHandle? frameworkHandle)
#if NETFRAMEWORK
        : this(sourceFileName, runSettings, frameworkHandle, new AppDomainWrapper())
#endif
    {
#if !WINDOWS_UWP && !NETFRAMEWORK
        _sourceFileName = sourceFileName;

        // Set the environment context.
        SetContext(sourceFileName);
#endif
    }

#if NETFRAMEWORK
    internal TestSourceHost(string sourceFileName, IRunSettings? runSettings, IFrameworkHandle? frameworkHandle, IAppDomain appDomain)
    {
        _sourceFileName = sourceFileName;
        _runSettings = runSettings;
        _frameworkHandle = frameworkHandle;
        _appDomain = appDomain;

        // Set the environment context.
        SetContext(sourceFileName);

        // Set isAppDomainCreationDisabled flag
        _isAppDomainCreationDisabled = _runSettings != null && MSTestAdapterSettings.IsAppDomainCreationDisabled(_runSettings.SettingsXml);
    }

    /// <summary>
    /// Gets the child AppDomain used to discover/execute tests.
    /// </summary>
    internal AppDomain? AppDomain { get; private set; }
#endif

    /// <summary>
    /// Setup the isolation host.
    /// </summary>
    public void SetupHost()
    {
#if NETFRAMEWORK
        List<string> resolutionPaths = GetResolutionPaths(_sourceFileName, VSInstallationUtilities.IsCurrentProcessRunningInPortableMode());

        if (EqtTrace.IsInfoEnabled)
        {
            EqtTrace.Info("DesktopTestSourceHost.SetupHost(): Creating assembly resolver with resolution paths {0}.", string.Join(",", resolutionPaths.ToArray()));
        }

        _parentDomainAssemblyResolver = new AssemblyResolver(resolutionPaths);
        AddSearchDirectoriesSpecifiedInRunSettingsToAssemblyResolver(_parentDomainAssemblyResolver, Path.GetDirectoryName(_sourceFileName));

        // Case when DisableAppDomain setting is present in runsettings and no child-appdomain needs to be created
        if (!_isAppDomainCreationDisabled)
        {
            // Setup app-domain
            var appDomainSetup = new AppDomainSetup();
            _targetFrameworkVersion = GetTargetFrameworkVersionString(_sourceFileName);
            AppDomainUtilities.SetAppDomainFrameworkVersionBasedOnTestSource(appDomainSetup, _targetFrameworkVersion);

            appDomainSetup.ApplicationBase = GetAppBaseAsPerPlatform();
            var configFile = GetConfigFileForTestSource(_sourceFileName);
            AppDomainUtilities.SetConfigurationFile(appDomainSetup, configFile);

            EqtTrace.Info("DesktopTestSourceHost.SetupHost(): Creating app-domain for source {0} with application base path {1}.", _sourceFileName, appDomainSetup.ApplicationBase);

            string domainName = string.Format("TestSourceHost: Enumerating source ({0})", _sourceFileName);
            AppDomain = _appDomain.CreateDomain(domainName, null!, appDomainSetup);

            // Load objectModel before creating assembly resolver otherwise in 3.5 process, we run into a recursive assembly resolution
            // which is trigged by AppContainerUtilities.AttachEventToResolveWinmd method.
            EqtTrace.SetupRemoteEqtTraceListeners(AppDomain);

            // Add an assembly resolver in the child app-domain...
            Type assemblyResolverType = typeof(AssemblyResolver);

            EqtTrace.Info("DesktopTestSourceHost.SetupHost(): assemblyenumerator location: {0} , fullname: {1} ", assemblyResolverType.Assembly.Location, assemblyResolverType.FullName);

            var resolver = AppDomainUtilities.CreateInstance(
                AppDomain,
                assemblyResolverType,
                new object[] { resolutionPaths });

            EqtTrace.Info(
                "DesktopTestSourceHost.SetupHost(): resolver type: {0} , resolve type assembly: {1} ",
                resolver.GetType().FullName,
                resolver.GetType().Assembly.Location);

            _childDomainAssemblyResolver = (AssemblyResolver)resolver;

            AddSearchDirectoriesSpecifiedInRunSettingsToAssemblyResolver(_childDomainAssemblyResolver, Path.GetDirectoryName(_sourceFileName));
        }
#endif
    }

    /// <summary>
    /// Creates an instance of a given type in the test source host.
    /// </summary>
    /// <param name="type"> The type that needs to be created in the host. </param>
    /// <param name="args">The arguments to pass to the constructor.
    /// This array of arguments must match in number, order, and type the parameters of the constructor to invoke.
    /// Pass in null for a constructor with no arguments.
    /// </param>
    /// <returns>  An instance of the type created in the host. </returns>
    /// <remarks> If a type is to be created in isolation then it needs to be a MarshalByRefObject. </remarks>
    public object? CreateInstanceForType(Type type, object?[]? args)
    {
#if NETFRAMEWORK
        // Honor DisableAppDomain setting if it is present in runsettings
        if (_isAppDomainCreationDisabled)
        {
            return Activator.CreateInstance(type, args);
        }

        return AppDomainUtilities.CreateInstance(AppDomain!, type, args);
#else
        return Activator.CreateInstance(type, args);
#endif
    }

    /// <summary>
    /// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
    /// </summary>
    public void Dispose()
    {
#if NETFRAMEWORK
        if (_parentDomainAssemblyResolver != null)
        {
            _parentDomainAssemblyResolver.Dispose();
            _parentDomainAssemblyResolver = null;
        }

        if (_childDomainAssemblyResolver != null)
        {
            _childDomainAssemblyResolver.Dispose();
            _childDomainAssemblyResolver = null;
        }

        if (AppDomain != null)
        {
            try
            {
                _appDomain.Unload(AppDomain);
            }
            catch (Exception exception)
            {
                // This happens usually when a test spawns off a thread and fails to clean it up.
                EqtTrace.Error("DesktopTestSourceHost.Dispose(): The app domain running tests could not be unloaded. Exception: {0}", exception);

                if (_frameworkHandle != null)
                {
                    // Let the test platform know that it should tear down the test host process
                    // since we have issues in unloading appdomain. We do so to avoid any assembly locking issues.
                    _frameworkHandle.EnableShutdownAfterTestRun = true;

                    EqtTrace.Verbose("DesktopTestSourceHost.Dispose(): Notifying the test platform that the test host process should be shut down because the app domain running tests could not be unloaded successfully.");
                }
            }

            AppDomain = null;
        }

        ResetContext();

        GC.SuppressFinalize(this);
#elif !WINDOWS_UWP
        ResetContext();
#endif
    }

#if !WINDOWS_UWP
    /// <summary>
    /// Sets context required for running tests.
    /// </summary>
    /// <param name="source">
    /// source parameter used for setting context.
    /// </param>
    private void SetContext(string? source)
    {
        if (StringEx.IsNullOrEmpty(source))
        {
            return;
        }

        Exception? setWorkingDirectoryException = null;
        _currentDirectory = Environment.CurrentDirectory;
        try
        {
            // If the source is in the format of an assembly qualified name, then calls to
            // Path.GetDirectoryName will return empty string. But if we use Path.GetFullPath first
            // then directory resolution works properly.
            var dirName = Path.GetDirectoryName(Path.GetFullPath(source));
#if WIN_UI
            if (StringEx.IsNullOrEmpty(dirName))
            {
                dirName = Path.GetDirectoryName(typeof(TestSourceHost).Assembly.Location)!;
            }

            Directory.SetCurrentDirectory(dirName);
#else
            Environment.CurrentDirectory = dirName!;
#if NETFRAMEWORK
            EqtTrace.Info("MSTestExecutor: Changed the working directory to {0}", Environment.CurrentDirectory);
#endif
#endif
        }
        catch (IOException ex)
        {
            setWorkingDirectoryException = ex;
        }
        catch (System.Security.SecurityException ex)
        {
            setWorkingDirectoryException = ex;
        }

        if (setWorkingDirectoryException != null)
        {
            EqtTrace.Error("MSTestExecutor.SetWorkingDirectory: Failed to set the working directory to '{0}'. {1}", Path.GetDirectoryName(source), setWorkingDirectoryException);
        }
    }

    /// <summary>
    /// Resets the context as it was before calling SetContext().
    /// </summary>
    private void ResetContext()
    {
        if (!StringEx.IsNullOrEmpty(_currentDirectory))
        {
            Environment.CurrentDirectory = _currentDirectory;
        }
    }
#endif

#if NETFRAMEWORK
    /// <summary>
    /// Gets child-domain's appbase to point to appropriate location.
    /// </summary>
    /// <returns>Appbase path that should be set for child appdomain.</returns>
    internal string GetAppBaseAsPerPlatform()
    {
        // The below logic of preferential setting the appdomains appbase is needed because:
        // 1. We set this to the location of the test source if it is built for Full CLR  -> Ideally this needs to be done in all situations.
        // 2. We set this to the location where the current adapter is being picked up from for UWP and .Net Core scenarios -> This needs to be
        //    different especially for UWP because we use the desktop adapter(from %temp%\VisualStudioTestExplorerExtensions) itself for test discovery
        //    in IDE scenarios. If the app base is set to the test source location, discovery will not work because we drop the
        //    UWP platform service assembly at the test source location and since CLR starts looking for assemblies from the app base location,
        //    there would be a mismatch of platform service assemblies during discovery.
        DebugEx.Assert(_targetFrameworkVersion is not null, "Target framework version is null.");
        if (_targetFrameworkVersion.Contains(Constants.DotNetFrameWorkStringPrefix))
        {
            return Path.GetDirectoryName(_sourceFileName) ?? Path.GetDirectoryName(typeof(TestSourceHost).Assembly.Location);
        }
        else
        {
            return Path.GetDirectoryName(typeof(TestSourceHost).Assembly.Location);
        }
    }

    /// <summary>
    /// Gets the probing paths to load the test assembly dependencies.
    /// </summary>
    /// <param name="sourceFileName">
    /// The source File Name.
    /// </param>
    /// <param name="isPortableMode">
    /// True if running in portable mode else false.
    /// </param>
    /// <returns>
    /// A list of path.
    /// </returns>
    internal virtual List<string> GetResolutionPaths(string sourceFileName, bool isPortableMode)
    {
        List<string> resolutionPaths = new()
        {
            // Add path of test assembly in resolution path. Mostly will be used for resolving winmd.
            Path.GetDirectoryName(sourceFileName),
        };

        if (!isPortableMode)
        {
            EqtTrace.Info("DesktopTestSourceHost.GetResolutionPaths(): Not running in portable mode");

            string? pathToPublicAssemblies = VSInstallationUtilities.PathToPublicAssemblies;
            if (!StringEx.IsNullOrWhiteSpace(pathToPublicAssemblies)
                && !StringUtilities.IsNullOrWhiteSpace(pathToPublicAssemblies))
            {
                resolutionPaths.Add(pathToPublicAssemblies);
            }

            string? pathToPrivateAssemblies = VSInstallationUtilities.PathToPrivateAssemblies;
            if (!StringEx.IsNullOrWhiteSpace(pathToPrivateAssemblies)
                && !StringUtilities.IsNullOrWhiteSpace(pathToPrivateAssemblies))
            {
                resolutionPaths.Add(pathToPrivateAssemblies);
            }
        }

        // Adding adapter folder to resolution paths
        if (!resolutionPaths.Contains(Path.GetDirectoryName(typeof(TestSourceHost).Assembly.Location)))
        {
            resolutionPaths.Add(Path.GetDirectoryName(typeof(TestSourceHost).Assembly.Location));
        }

        // Adding TestPlatform folder to resolution paths
        if (!resolutionPaths.Contains(Path.GetDirectoryName(typeof(AssemblyHelper).Assembly.Location)))
        {
            resolutionPaths.Add(Path.GetDirectoryName(typeof(AssemblyHelper).Assembly.Location));
        }

        return resolutionPaths;
    }

    internal virtual string GetTargetFrameworkVersionString(string sourceFileName)
    {
        return AppDomainUtilities.GetTargetFrameworkVersionString(sourceFileName);
    }

    private static string? GetConfigFileForTestSource(string sourceFileName)
    {
        return new DeploymentUtility().GetConfigFile(sourceFileName);
    }

    private static void AddSearchDirectoriesSpecifiedInRunSettingsToAssemblyResolver(AssemblyResolver assemblyResolver, string baseDirectory)
    {
        // Check if user specified any adapter settings
        MSTestAdapterSettings adapterSettings = MSTestSettingsProvider.Settings;
        if (adapterSettings == null)
        {
            return;
        }

        try
        {
            var additionalSearchDirectories = adapterSettings.GetDirectoryListWithRecursiveProperty(baseDirectory);
            if (additionalSearchDirectories?.Count > 0)
            {
                assemblyResolver.AddSearchDirectoriesFromRunSetting(additionalSearchDirectories);
            }
        }
        catch (Exception exception)
        {
            EqtTrace.Error(
                "DesktopTestSourceHost.AddSearchDirectoriesSpecifiedInRunSettingsToAssemblyResolver(): Exception hit while trying to set assembly resolver for domain. Exception : {0} \n Message : {1}",
                exception,
                exception.Message);
        }
    }
#endif
}
